from thunter.cli import thunter_cli_app
from thunter.tests import CliCommandTestBaseClass


class TestList(CliCommandTestBaseClass):
    def test_list_tasks(self):
        result = self.runner.invoke(thunter_cli_app, ["ls"])
        self.assertIn("Current", result.output)
        self.assertIn("In Progress", result.output)
        self.assertIn("TODO", result.output)
        self.assertNotIn("Finished", result.output)

    def test_list_all_tasks(self):
        result = self.runner.invoke(thunter_cli_app, ["ls", "--all"])
        self.assertIn("Current", result.output)
        self.assertIn("In Progress", result.output)
        self.assertIn("TODO", result.output)
        self.assertIn("Finished", result.output)

    def test_list_finished_tasks(self):
        result = self.runner.invoke(thunter_cli_app, ["ls", "--finished"])
        self.assertNotIn("Current", result.output)
        self.assertNotIn("In Progress", result.output)
        self.assertNotIn("TODO", result.output)
        self.assertIn("Finished", result.output)

    def test_list_todo_tasks(self):
        result = self.runner.invoke(thunter_cli_app, ["ls", "--todo"])
        self.assertNotIn("Current", result.output)
        self.assertNotIn("In Progress", result.output)
        self.assertIn("TODO", result.output)
        self.assertNotIn("Finished", result.output)

    def test_list_open_tasks(self):
        result = self.runner.invoke(thunter_cli_app, ["ls", "--open"])
        self.assertIn("Current", result.output)
        self.assertIn("In Progress", result.output)
        self.assertIn("TODO", result.output)
        self.assertNotIn("Finished", result.output)

    def test_list_started_tasks(self):
        result = self.runner.invoke(thunter_cli_app, ["ls", "--started"])
        self.assertIn("Current", result.output)
        self.assertIn("In Progress", result.output)
        self.assertNotIn("TODO", result.output)
        self.assertNotIn("Finished", result.output)

    def test_list_current_tasks(self):
        result = self.runner.invoke(thunter_cli_app, ["ls", "--current"])
        self.assertIn("Current", result.output)
        self.assertNotIn("In Progress", result.output)
        self.assertNotIn("TODO", result.output)
        self.assertNotIn("Finished", result.output)

    def test_list_in_progress_tasks(self):
        result = self.runner.invoke(thunter_cli_app, ["ls", "--in-progress"])
        self.assertNotIn("Current", result.output)
        self.assertIn("In Progress", result.output)
        self.assertNotIn("TODO", result.output)
        self.assertNotIn("Finished", result.output)

    def test_list_starts_with(self):
        result = self.runner.invoke(thunter_cli_app, ["ls", "--starts-with", "a "])
        self.assertNotIn("another great test task", result.output)
        self.assertIn("a test task", result.output)
        self.assertIn("a long task", result.output)
        self.assertNotIn("a finished task", result.output)

    def test_list_contains(self):
        result = self.runner.invoke(thunter_cli_app, ["ls", "--contains", "test"])
        self.assertIn("another great test task", result.output)
        self.assertIn("a test task", result.output)
        self.assertNotIn("a finished task", result.output)
        self.assertNotIn("a long task", result.output)



from dataclasses import dataclass
from time import time
from functools import total_ordering

from thunter.time import display_time


@total_ordering
@dataclass
class TaskHistoryRecord:
    """Represents a task history record, used to track time spent working on the task."""

    id: int
    taskid: int
    is_start: bool
    time: int

    @classmethod
    def from_db_record(cls, record: tuple[int, int, bool, int]):
        return cls(
            id=record[0], taskid=record[1], is_start=bool(record[2]), time=record[3]
        )

    @classmethod
    def calc_progress(cls, task_history: list["TaskHistoryRecord"]) -> int:
        """Calculates the total time (seconds) spent on a task based on its history records."""
        progress = 0
        start_time = 0
        for history_record in sorted(task_history):
            if history_record.is_start:
                start_time = history_record.time
            else:
                progress += history_record.time - start_time
                start_time = 0
        if task_history and start_time:
            progress += int(time()) - start_time
        return progress

    @classmethod
    def display_progress(cls, seconds: int) -> str:
        """Formats the progress time (int seconds) in HH:MM:SS format."""
        minutes, seconds = divmod(seconds, 60)
        hours, minutes = divmod(minutes, 60)
        return "{0:02d}:{1:02d}:{2:02d}".format(hours, minutes, seconds)

    @property
    def time_display(self) -> str:
        return display_time(self.time)

    def __lt__(self, other: "TaskHistoryRecord"):
        return (self.taskid, self.time, not self.is_start) < (
            other.taskid,
            other.time,
            not other.is_start,
        )

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id



from typing import Annotated
import typer

from thunter.settings import thunter_print
from thunter.task_hunter import TaskHunter


app = typer.Typer()


@app.command("rm, remove")
def rm(
    task_id: Annotated[
        list[str], typer.Argument(help="ID or name of task to be removed.")
    ],
    force: Annotated[
        bool, typer.Option("--force", "-f", help="No confirmation prompt")
    ] = False,
):
    """Remove/delete a task."""
    hunter = TaskHunter()
    task_identifier = " ".join(task_id)
    task = hunter.get_task(task_identifier)

    if force:
        user_is_sure = True
    else:
        prompt = (
            f"Are you sure you want to permanently delete [red]{task.name}[/red]!? [yN]"
        )
        user_is_sure = input(prompt).lower() == "y"

    if user_is_sure:
        hunter.remove_task(task.id)
        thunter_print(f"Removed [red]{task.name}[/red]!")
    else:
        thunter_print(f"Didn't remove [yellow]{task.name}[/yellow].")



import os
from rich.console import Console


THUNTER_DIR = os.path.expanduser(os.environ.get("THUNTER_DIRECTORY", "~/.thunter"))
DATABASE = os.path.join(
    THUNTER_DIR, os.environ.get("THUNTER_DATABASE_NAME", "thunter_database.db")
)
EDITOR = os.environ.get("EDITOR", "vim")
THUNTER_SILENT = os.environ.get("THUNTER_SILENT", "false").lower() in (
    "true",
    "1",
    "yes",
    "y",
)
DEBUG = os.environ.get("DEBUG", "false").lower() in ("true", "1", "yes", "y")

# Global configuration for printing, updatable by CLI options and initialized
# from environment variables or defaults above.
print_config = {"silent": THUNTER_SILENT, "debug": DEBUG}


def thunter_print(*args, **kwargs):
    """Prints to console if not in silent mode."""
    if not THUNTER_SILENT and not print_config["silent"]:
        console = Console()
        console.print(*args, **kwargs)


def needs_init():
    """Checks if `thunter init` needs to be run to setup the environment."""
    return (
        not THUNTER_DIR
        or not DATABASE
        or not os.path.exists(THUNTER_DIR)
        or not os.path.exists(DATABASE)
    )



import re
import sys
from typing_extensions import Annotated

import typer.core
from rich.console import Console

from thunter import settings
from thunter.cli.create import app as create_app
from thunter.cli.db import app as db_app
from thunter.cli.edit import app as edit_app
from thunter.cli.estimate import app as estimate_app
from thunter.cli.finish import app as finish_app
from thunter.cli.init import app as init_app, init
from thunter.cli.ls import app as ls_app
from thunter.cli.restart import app as restart_app
from thunter.cli.rm import app as rm_app
from thunter.cli.show import app as show_app
from thunter.cli.stop import app as stop_app
from thunter.cli.workon import app as workon_app
from thunter.constants import ThunterError


class AliasGroup(typer.core.TyperGroup):
    """Custom override of TyperGroup to support command aliases.

    Watch this issue for possible native support in Typer for aliases:
    https://github.com/fastapi/typer/issues/1242
    """

    _CMD_SPLIT_P = re.compile(r" ?[,|] ?")

    def get_command(self, ctx, cmd_name):
        cmd_name = self._group_cmd_name(cmd_name)
        return super().get_command(ctx, cmd_name)

    def _group_cmd_name(self, default_name):
        for cmd in self.commands.values():
            name = cmd.name
            if name and default_name in self._CMD_SPLIT_P.split(name):
                return name
        return default_name


thunter_cli_app = typer.Typer(
    name="thunter",
    no_args_is_help=True,
    cls=AliasGroup,
)
# Add all subcommands to the main CLI app
thunter_cli_app.add_typer(init_app)
thunter_cli_app.add_typer(ls_app)
thunter_cli_app.add_typer(show_app)
thunter_cli_app.add_typer(workon_app)
thunter_cli_app.add_typer(create_app)
thunter_cli_app.add_typer(restart_app)
thunter_cli_app.add_typer(stop_app)
thunter_cli_app.add_typer(finish_app)
thunter_cli_app.add_typer(estimate_app)
thunter_cli_app.add_typer(edit_app)
thunter_cli_app.add_typer(rm_app)
thunter_cli_app.add_typer(db_app)


@thunter_cli_app.callback()
def main_callback(
    ctx: typer.Context,
    silent: Annotated[
        bool,
        typer.Option(
            "--silent",
            "--quite",
            "-s",
            help="Run thunter in silent mode, no output to console.",
        ),
    ] = False,
    debug: Annotated[
        bool,
        typer.Option(
            "--debug",
            "-d",
            help="Run thunter in debug mode, printing out full exceptions and traces.",
        ),
    ] = False,
):
    """THunter - your task hunter, tracking time spent on your TODO list!"""
    settings.print_config["silent"] = silent or settings.THUNTER_SILENT
    settings.print_config["debug"] = debug or settings.DEBUG
    if ctx.invoked_subcommand != "init" and settings.needs_init():
        ctx.invoke(init)


def main():
    try:
        thunter_cli_app()
    except KeyboardInterrupt:
        sys.exit(1)
    except ThunterError as thunter_error:
        console = Console()
        if settings.print_config["debug"]:
            console.print_exception(show_locals=True)
        console.print(str(thunter_error))
        sys.exit(thunter_error.exit_status)


if __name__ == "__main__":
    main()



from thunter.tests import CliCommandTestBaseClass
from thunter.cli import thunter_cli_app


class TestShow(CliCommandTestBaseClass):
    def test_show_task(self):
        result = self.runner.invoke(thunter_cli_app, ["show"])
        self.assertIn("a long task", result.output)
        self.assertIn("Current", result.output)

        result = self.runner.invoke(thunter_cli_app, ["show", "a test task"])
        self.assertIn("a test task", result.output)

        result = self.runner.invoke(thunter_cli_app, ["show", "4"])
        self.assertIn("a finished task", result.output)

    def test_show_nonexistent_task(self):
        result = self.runner.invoke(thunter_cli_app, ["show", "999"])
        self.assertIn("Could not find task for identifier: 999", str(result.exception))
        self.assertGreater(result.exit_code, 0)

    def test_show_most_recent_task(self):
        self.thunter.stop_current_task()
        result = self.runner.invoke(thunter_cli_app, ["show"])
        self.assertIn("a long task", str(result.output))



from thunter.tests import CliCommandTestBaseClass
from thunter.cli import thunter_cli_app


class TestFinish(CliCommandTestBaseClass):
    def test_finish_current_task(self):
        result = self.runner.invoke(thunter_cli_app, ["finish"])
        self.assertIn("Finished a long task", result.output)

        result = self.runner.invoke(thunter_cli_app, ["show", "a finished task"])
        self.assertIn("Finished", result.output)

    def test_finish_specific_test(self):
        result = self.runner.invoke(thunter_cli_app, ["finish", "6"])
        self.assertIn("Finished identically named task", result.output)

        result = self.runner.invoke(thunter_cli_app, ["show", "6"])
        self.assertIn("Finished", result.output)

    def test_finish_no_task_found(self):
        result = self.runner.invoke(thunter_cli_app, ["finish", "9999"])
        self.assertIn("Could not find task for identifier: 9999", str(result.exception))
        self.assertGreater(result.exit_code, 0)



from thunter.tests import CliCommandTestBaseClass
from thunter.cli import thunter_cli_app


class TestEstimate(CliCommandTestBaseClass):
    def test_estimate_current_task(self):
        result = self.runner.invoke(thunter_cli_app, ["estimate", "7"])
        self.assertIn("a long task estimated to take 7 hrs.", result.output)

    def test_estimate_no_current_task(self):
        self.runner.invoke(thunter_cli_app, ["stop"])
        result = self.runner.invoke(thunter_cli_app, ["estimate", "7"])
        self.assertIn("No current task found to estimate.", str(result.exception))
        self.assertGreater(result.exit_code, 0)

    def test_estimate_specific_tas(self):
        result = self.runner.invoke(
            thunter_cli_app, ["estimate", "--task-identifier", "a test task", "7"]
        )
        self.assertIn("a test task estimated to take 7 hrs.", result.output)



from dataclasses import dataclass
import os
import shutil
import sqlite3
import tempfile
from unittest import TestCase

from typer.testing import CliRunner

from thunter import settings
from thunter.task_hunter import TaskHunter


@dataclass
class TestDatabaseEnvironment:
    THUNTER_DIR: str
    DATABASE: str


def setUpTestDatabase() -> TestDatabaseEnvironment:
    """Setup a temporary environment and database for testing."""
    thunter_dir = tempfile.mkdtemp()
    env = TestDatabaseEnvironment(
        THUNTER_DIR=thunter_dir,
        DATABASE=os.path.join(thunter_dir, "test_database.db"),
    )
    settings.THUNTER_DIR = env.THUNTER_DIR
    settings.DATABASE = env.DATABASE

    conn = sqlite3.connect(env.DATABASE)
    dir_path = os.path.dirname(os.path.realpath(__file__))
    with open(dir_path + "/test_database_fixture.sql", "r") as f:
        conn.executescript(f.read())
    conn.close()

    return env


def tearDownTestDatabase(env: TestDatabaseEnvironment) -> None:
    """Remove the temporary environment and database after tests."""
    shutil.rmtree(env.THUNTER_DIR)


class CliCommandTestBaseClass(TestCase):
    def setUp(self):
        self.env = setUpTestDatabase()
        self.thunter = TaskHunter()
        self.runner = CliRunner()

    def tearDown(self):
        """Remove the temporary environment and database after tests."""
        tearDownTestDatabase(self.env)



from unittest import TestCase

from thunter.constants import Status, ThunterTaskValidationError
from thunter.models import Task, TaskHistoryRecord
from thunter.parser import (
    ParsedTaskData,
    ParsedTaskHistoryRecord,
    display_task,
    parse_task_display,
    validate_task_data,
    name,
    estimate,
    status,
    description,
    history,
)


class TestTaskParser(TestCase):
    def test_parse_task(self):
        """Test parsing a task from a string."""
        task = Task(
            id=1,
            name="Test Task",
            description="This is a test task.",
            estimate=4,
            status=Status.IN_PROGRESS,
            last_modified_at=1633036800,
            created_at=1633036800,
        )
        task_history = [
            TaskHistoryRecord(
                id=1,
                taskid=1,
                is_start=True,
                time=1633036800,
            ),
            TaskHistoryRecord(
                id=2,
                taskid=1,
                is_start=False,
                time=1633037200,
            ),
        ]
        task_display = display_task(task, task_history)
        expected_display = (
            "NAME: Test Task\n"
            "ESTIMATE: 4\n"
            "STATUS: In Progress\n"
            "DESCRIPTION: This is a test task.\n"
            "\n"
            "HISTORY\n"
            "Start\t2021-09-30 21:20:00\n"
            "Stop\t2021-09-30 21:26:40\n"
        )
        self.assertEqual(task_display, expected_display)

        parsed_data = parse_task_display(task_display)

        self.assertEqual(parsed_data.name, "Test Task")
        self.assertEqual(parsed_data.estimate, 4)
        self.assertEqual(parsed_data.description, "This is a test task.")
        self.assertEqual(parsed_data.status, Status.IN_PROGRESS)
        self.assertEqual(len(parsed_data.history), 2)
        self.assertTrue(parsed_data.history[0].is_start)
        self.assertEqual(parsed_data.history[0].time, 1633036800)
        self.assertFalse(parsed_data.history[1].is_start)
        self.assertEqual(parsed_data.history[1].time, 1633037200)

        task_without_history_display = (
            "NAME: Test Task\n"
            "ESTIMATE: 4\n"
            "STATUS: TODO\n"
            "DESCRIPTION: This is a test task.\n"
            "\n"
            "HISTORY\n"
        )

        parsed_data_without_history = parse_task_display(task_without_history_display)

        self.assertEqual(parsed_data_without_history.name, "Test Task")
        self.assertEqual(parsed_data_without_history.estimate, 4)
        self.assertEqual(
            parsed_data_without_history.description, "This is a test task."
        )
        self.assertEqual(parsed_data_without_history.status, Status.TODO)
        self.assertEqual(len(parsed_data_without_history.history), 0)

    def test_validate_task_data(self):
        """Test validating task data."""
        task_data = ParsedTaskData(
            name="TODO task with history!",
            estimate=3,
            description="A valid task for testing.",
            status=Status.TODO,
            history=[
                ParsedTaskHistoryRecord(is_start=True, time=1633036800),
                ParsedTaskHistoryRecord(is_start=False, time=1633037200),
            ],
        )
        with self.assertRaises(ThunterTaskValidationError) as error:
            validate_task_data(task_data)
        self.assertIn(
            "Can't have a history if the status is TODO", str(error.exception)
        )

        task_data = ParsedTaskData(
            name="IN_PROGRESS task without history!",
            estimate=3,
            description=None,
            status=Status.IN_PROGRESS,
            history=[],
        )
        with self.assertRaises(ThunterTaskValidationError) as error:
            validate_task_data(task_data)
        self.assertIn(
            "Must have a history if status is In Progress", str(error.exception)
        )

        task_data = ParsedTaskData(
            name="CURRENT task that isn't tracking time!",
            estimate=3,
            description=None,
            status=Status.CURRENT,
            history=[
                ParsedTaskHistoryRecord(is_start=True, time=1633036800),
                ParsedTaskHistoryRecord(is_start=False, time=1633037200),
            ],
        )
        with self.assertRaises(ThunterTaskValidationError) as error:
            validate_task_data(task_data)
        self.assertIn(
            "Last history record must be a Start if the status is Current",
            str(error.exception),
        )

        task_data = ParsedTaskData(
            name="IN_PROGRESS task that is tracking time!",
            estimate=3,
            description=None,
            status=Status.IN_PROGRESS,
            history=[
                ParsedTaskHistoryRecord(is_start=True, time=1633036800),
            ],
        )
        with self.assertRaises(ThunterTaskValidationError) as error:
            validate_task_data(task_data)
        self.assertIn(
            "Last history record must be a Stop if the status is In Progress",
            str(error.exception),
        )

        task_data = ParsedTaskData(
            name="FINISHED task that is tracking time!",
            estimate=3,
            description=None,
            status=Status.FINISHED,
            history=[
                ParsedTaskHistoryRecord(is_start=True, time=1633036800),
            ],
        )
        with self.assertRaises(ThunterTaskValidationError) as error:
            validate_task_data(task_data)
        self.assertIn(
            "Last history record must be a Stop if the status is Finished",
            str(error.exception),
        )

        task_data = ParsedTaskData(
            name="Out of order history",
            estimate=3,
            description=None,
            status=Status.FINISHED,
            history=[
                ParsedTaskHistoryRecord(is_start=True, time=1633036805),
                ParsedTaskHistoryRecord(is_start=False, time=1633036800),
            ],
        )
        with self.assertRaises(ThunterTaskValidationError) as error:
            validate_task_data(task_data)
        self.assertIn(
            "History must be in ascending order by time",
            str(error.exception),
        )

        task_data = ParsedTaskData(
            name="Stop and Start reversed",
            estimate=3,
            description=None,
            status=Status.FINISHED,
            history=[
                ParsedTaskHistoryRecord(is_start=True, time=1633036800),
                ParsedTaskHistoryRecord(is_start=True, time=1633036800),
                ParsedTaskHistoryRecord(is_start=False, time=1633036800),
                ParsedTaskHistoryRecord(is_start=False, time=1633036800),
            ],
        )
        with self.assertRaises(ThunterTaskValidationError) as error:
            validate_task_data(task_data)
        self.assertIn(
            "History must alternate between Start and Stop",
            str(error.exception),
        )

    def test_parse_name(self):
        """Test parsing a task name."""
        name_display = "NAME: Test Task"
        parsed_data = name.parse_string(name_display)
        self.assertEqual(parsed_data.name, "Test Task")

    def test_parse_estimate(self):
        """Test parsing a task estimate."""
        estimate_display = "ESTIMATE: 5"
        parsed_data = estimate.parse_string(estimate_display)
        self.assertEqual(parsed_data.estimate, 5)

    def test_parse_status(self):
        """Test parsing a task status."""
        status_display = "STATUS: In Progress"
        parsed_data = status.parse_string(status_display)
        self.assertEqual(parsed_data.status, Status.IN_PROGRESS)

    def test_parse_description(self):
        """Test parsing a task description."""
        description_display = "DESCRIPTION: This is a test task."
        parsed_data = description.parse_string(description_display)
        self.assertEqual(parsed_data.description, "This is a test task.")

    def test_parse_empty_description(self):
        """Test parsing a task description."""
        description_display = "DESCRIPTION: "
        parsed_data = description.parse_string(description_display)
        self.assertEqual(parsed_data.description, "")

    def test_parse_history(self):
        """Test parsing task history."""
        history_display = (
            "HISTORY\nStart\t2021-09-30 21:20:00\nStop\t2021-09-30 21:26:40\n"
        )
        parsed_data = history.parse_string(history_display)
        self.assertEqual(len(parsed_data.history), 2)
        self.assertTrue(parsed_data.history[0].is_start)  # type: ignore
        self.assertEqual(parsed_data.history[0].time, 1633036800)  # type: ignore
        self.assertFalse(parsed_data.history[1].is_start)  # type: ignore
        self.assertEqual(parsed_data.history[1].time, 1633037200)  # type: ignore



from typing import Annotated
import typer

from thunter.cli.ls import ls
from thunter.constants import Status
from thunter.task_hunter import TaskHunter


app = typer.Typer()


@app.command()
def restart(ctx: typer.Context, task_id: Annotated[str, typer.Argument()]):
    """Restart a finished task (progress will continue from before)."""
    hunter = TaskHunter()
    task = hunter.get_task(task_id, statuses=set([Status.FINISHED]))
    hunter.workon_task(task.id)
    ctx.invoke(ls, current=True)



from thunter.tests import CliCommandTestBaseClass
from thunter.cli import thunter_cli_app


class TestStop(CliCommandTestBaseClass):
    def test_stop(self):
        result = self.runner.invoke(thunter_cli_app, ["stop"])
        self.assertIn("Stopped working on a long task", result.output)

        result = self.runner.invoke(thunter_cli_app, ["stop"])
        self.assertIn("No current task to stop", result.output)



from collections import defaultdict
from typing import Annotated

import typer
from rich import box
from rich.table import Table

from thunter.constants import Status
from thunter.models.task_history_record import TaskHistoryRecord
from thunter.settings import thunter_print
from thunter.task_hunter import TaskHunter


app = typer.Typer()


@app.command("ls, list")
def ls(
    all: Annotated[
        bool | None,
        typer.Option(
            "--all",
            "-a",
            help="List all tasks (short for -citf)",
            rich_help_panel="Status Filtering",
        ),
    ] = None,
    open: Annotated[
        bool | None,
        typer.Option(
            "--open",
            "-o",
            help="List all open tasks (short for -cit)",
            rich_help_panel="Status Filtering",
        ),
    ] = None,
    started: Annotated[
        bool | None,
        typer.Option(
            "--started",
            "-s",
            help="List all started tasks (short for -ci)",
            rich_help_panel="Status Filtering",
        ),
    ] = None,
    current: Annotated[
        bool | None,
        typer.Option(
            "--current",
            "-c",
            help="List all Current tasks",
            rich_help_panel="Status Filtering",
        ),
    ] = None,
    in_progress: Annotated[
        bool | None,
        typer.Option(
            "--in-progress",
            "-i",
            help="List all In Progress tasks",
            rich_help_panel="Status Filtering",
        ),
    ] = None,
    todo: Annotated[
        bool | None,
        typer.Option(
            "--todo",
            "-t",
            help="List all TODO tasks",
            rich_help_panel="Status Filtering",
        ),
    ] = None,
    finished: Annotated[
        bool | None,
        typer.Option(
            "--finished",
            "-f",
            help="List all Finished tasks",
            rich_help_panel="Status Filtering",
        ),
    ] = None,
    starts_with: Annotated[
        str | None,
        typer.Option(
            "--starts-with",
            "-S",
            help="Only tasks that start with STRING",
            rich_help_panel="Task Name Filtering",
        ),
    ] = None,
    contains: Annotated[
        str | None,
        typer.Option(
            "--contains",
            "-C",
            help="Only tasks that contain STRING",
            rich_help_panel="Task Name Filtering",
        ),
    ] = None,
):
    """List tasks. Defaults to listing all open tasks (CURRENT, IN_PROGRESS, TODO)."""
    statuses: set[Status] = set()
    if all:
        statuses.update(Status)
    if open:
        statuses.update([Status.CURRENT, Status.IN_PROGRESS, Status.TODO])
    if started:
        statuses.update([Status.CURRENT, Status.IN_PROGRESS])
    if current:
        statuses.add(Status.CURRENT)
    if in_progress:
        statuses.add(Status.IN_PROGRESS)
    if todo:
        statuses.add(Status.TODO)
    if finished:
        statuses.add(Status.FINISHED)
    if not statuses:
        statuses.update([Status.CURRENT, Status.IN_PROGRESS, Status.TODO])

    # Get the filtered and sorted list of tasks to display
    hunter = TaskHunter()
    tasks = hunter.get_tasks(
        statuses,
        starts_with=starts_with,
        contains=contains,
    )

    # Calculate progress from history
    history_records = hunter.get_history([task.id for task in tasks])
    taskid2history = defaultdict(list)
    for record in history_records:
        taskid2history[record.taskid].append(record)
    taskid2progress = defaultdict(int)
    for taskid, task_history in taskid2history.items():
        taskid2progress[taskid] = TaskHistoryRecord.calc_progress(task_history)

    # Pretty display in a table with colors
    table = Table(
        "ID",
        "NAME",
        "ESTIMATE",
        "PROGRESS",
        "STATUS",
        box=box.MINIMAL_HEAVY_HEAD,
    )
    for task in tasks:
        row = (
            str(task.id),
            task.name,
            task.estimate_display,
            TaskHistoryRecord.display_progress(taskid2progress[task.id]),
            task.status.value,
        )
        style = None
        if task.status == Status.CURRENT:
            style = "yellow"
        elif task.status == Status.IN_PROGRESS:
            style = "orange3"
        elif task.status == Status.FINISHED:
            style = "green"
        table.add_row(*row, style=style)
    thunter_print(table)



from typing import Annotated
import typer

from thunter.cli.show import show
from thunter.task_hunter import TaskHunter


app = typer.Typer()


@app.command()
def create(
    ctx: typer.Context,
    task_id: Annotated[list[str], typer.Argument()],
    estimate: Annotated[
        int | None,
        typer.Option(
            "--estimate",
            "-e",
            help="Add estimate (in hours)",
            prompt="Estimate (hours)",
            min=1,
        ),
    ] = None,
    description: Annotated[
        str | None, typer.Option("--description", "-d", help="Add a description")
    ] = None,
):
    """Create a new task."""
    hunter = TaskHunter()
    new_task = hunter.create_task(
        name=" ".join(task_id),
        estimate=estimate,
        description=description,
    )
    ctx.invoke(show, task_id=str(new_task.id))



from dataclasses import dataclass
from functools import total_ordering

from thunter.constants import (
    STATUS_ORDERING,
    Status,
)
from thunter.time import display_time


# task identifiers can be their ID or their name
TaskIdentifier = int | str


@total_ordering
@dataclass
class Task:
    """Represents a task in the THunter application."""

    id: int
    name: str
    estimate: int | None
    description: str | None
    status: Status
    last_modified_at: int
    created_at: int

    @classmethod
    def from_db_record(
        cls, record: tuple[int, str, int | None, str | None, str, int, int]
    ):
        return cls(
            id=record[0],
            name=record[1],
            estimate=record[2],
            description=record[3],
            status=Status(record[4]),
            last_modified_at=record[5],
            created_at=record[6],
        )

    @property
    def last_modified_at_display(self):
        return display_time(self.last_modified_at)

    @property
    def estimate_display(self):
        estimate_display_str = ""
        if self.estimate is not None:
            estimate_display_str = "%d hr" % self.estimate
            if self.estimate > 1:
                estimate_display_str += "s"
        return estimate_display_str

    def __lt__(self, other):
        return (STATUS_ORDERING.index(self.status.value), -self.last_modified_at) < (
            STATUS_ORDERING.index(other.status.value),
            -other.last_modified_at,
        )

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id



import os
import tempfile
from unittest import TestCase

from typer.testing import CliRunner

from thunter import settings
from thunter.cli import thunter_cli_app


class TestMainCallback(TestCase):
    def test_help(self):
        runner = CliRunner()
        result = runner.invoke(thunter_cli_app, ["--help"])
        self.assertIn("Usage: thunter", result.output)
        self.assertIn("THunter - your task hunter", result.output)

    def test_init_called_by_default(self):
        with tempfile.TemporaryDirectory() as thunter_dir:
            settings.THUNTER_DIR = thunter_dir
            settings.DATABASE = os.path.join(thunter_dir, "uninitialized.db")
            runner = CliRunner()
            result = runner.invoke(thunter_cli_app, ["ls"])
            self.assertIn("Initializing THunter...", result.output)

            self.assertTrue(os.path.exists(thunter_dir + "/uninitialized.db"))

    def test_thunter_silent(self):
        with tempfile.TemporaryDirectory() as thunter_dir:
            settings.THUNTER_DIR = thunter_dir
            settings.DATABASE = os.path.join(thunter_dir, "uninitialized.db")
            runner = CliRunner()
            result = runner.invoke(
                thunter_cli_app,
                ["--silent", "create", "Silent Task", "--estimate", "1"],
            )
            self.assertEqual("", result.output)



from unittest import TestCase

from thunter.models import TaskHistoryRecord


class TestTaskHistoryRecord(TestCase):
    def test_from_db_record(self):
        """Test creating a task from a database record."""
        record = (1, 2, True, 1633036800)
        task = TaskHistoryRecord.from_db_record(record)
        self.assertEqual(task.id, 1)
        self.assertEqual(task.taskid, 2)
        self.assertEqual(task.is_start, True)
        self.assertEqual(task.time, 1633036800)
        self.assertEqual(task.time_display, "2021-09-30 21:20:00")

    def test_calc_progress(self):
        """Test calculating progress from task history records."""
        history = [
            TaskHistoryRecord(id=2, taskid=1, is_start=False, time=1633021171),
            TaskHistoryRecord(id=4, taskid=1, is_start=False, time=1633021260),
            TaskHistoryRecord(id=1, taskid=1, is_start=True, time=1633021111),
            TaskHistoryRecord(id=3, taskid=1, is_start=True, time=1633021200),
        ]
        progress = TaskHistoryRecord.calc_progress(history)
        self.assertEqual(progress, 120)

    def test_display_progress(self):
        """Test displaying progress in HH:MM:SS format."""
        seconds = 3661
        formatted_progress = TaskHistoryRecord.display_progress(seconds)
        self.assertEqual(formatted_progress, "01:01:01")

    def test_ordering(self):
        """Test ordering of tasks based on statuses and last modified time."""
        history_record1 = TaskHistoryRecord(
            id=1,
            taskid=1,
            is_start=True,
            time=1633021111,
        )
        history_record2 = TaskHistoryRecord(
            id=2,
            taskid=1,
            is_start=False,
            time=1633021171,
        )
        history_record3 = TaskHistoryRecord(
            id=3,
            taskid=2,
            is_start=True,
            time=1633021111,
        )
        history_record4 = TaskHistoryRecord(
            id=4,
            taskid=2,
            is_start=False,
            time=1633021111,
        )

        sorted_history = sorted(
            [history_record4, history_record3, history_record2, history_record1]
        )
        self.assertEqual(
            sorted_history,
            [history_record1, history_record2, history_record3, history_record4],
        )



import pandas as pd

from thunter.constants import TableName
from thunter.db import Database


class TaskAnalyzer(Database):
    """Collection data analysis operations for tasks and history."""

    def fetch_data_df(self) -> pd.DataFrame:
        """Fetch the data as a DataFrame for use by analysis tools."""
        with self.connect() as con:
            df = pd.read_sql(
                (
                    f"SELECT tasks.id, tasks.estimate, tasks.name, history.is_start, history.time FROM {TableName.TASKS.value} "
                    f"INNER JOIN {TableName.HISTORY.value} ON {TableName.TASKS.value}.id = {TableName.HISTORY.value}.taskid "
                    f"ORDER BY history.time ASC"
                ),
                con=con,
                dtype={
                    "estimate": pd.Int32Dtype(),
                    "is_start": pd.BooleanDtype(),
                    "name": pd.StringDtype(),
                    "time": pd.Int64Dtype(),
                },
            )

        return df

    def fetch_tasks_df(self) -> pd.DataFrame:
        """Fetch the tasks as a DataFrame."""
        with self.connect() as con:
            df = pd.read_sql(
                f"SELECT id, estimate, name, status, created_at FROM {TableName.TASKS.value}",
                con=con,
                parse_dates={"created_at": "s"},
            )

        return df

    def fetch_history_df(self) -> pd.DataFrame:
        """Fetch the tasks as a DataFrame."""
        with self.connect() as con:
            df = pd.read_sql(
                f"SELECT id, taskid, is_start, time FROM {TableName.HISTORY.value} ORDER BY taskid, time ASC",
                con=con,
                dtype={"is_start": bool},
            )

        return df



from typing import Annotated
import typer

from thunter.cli.ls import ls
from thunter.constants import Status, ThunterCouldNotFindTaskError
from thunter.task_hunter import TaskHunter


app = typer.Typer()


@app.command("w, workon")
def workon(
    ctx: typer.Context,
    task_id: Annotated[list[str] | None, typer.Argument()] = None,
    create: Annotated[
        bool | None,
        typer.Option(
            "--create",
            "-c",
            help="Create the task if it does not exist",
            rich_help_panel="Task Creation",
        ),
    ] = None,
    estimate_hours: Annotated[
        int | None,
        typer.Option(
            "--estimate",
            "-e",
            help="Add estimate (in hours) when creating a task",
            min=1,
            rich_help_panel="Task Creation",
        ),
    ] = None,
    description: Annotated[
        str | None,
        typer.Option(
            "--description",
            "-d",
            help="Add a description when creating a task",
            rich_help_panel="Task Creation",
        ),
    ] = None,
):
    """Start/continue working on an unfinished task."""
    hunter = TaskHunter()
    task_id_str = " ".join(task_id) if task_id else None
    try:
        task = hunter.get_task(
            task_identifier=task_id_str,
            statuses=set([Status.CURRENT, Status.IN_PROGRESS, Status.TODO]),
            exact_match=True,
        )
    except ThunterCouldNotFindTaskError:
        if task_id_str:
            if create:
                task = hunter.create_task(
                    name=task_id_str,
                    estimate=get_estimate(estimate_hours),
                    description=description,
                )
            else:
                # try to find task with a partial match
                task = hunter.get_task(
                    task_identifier=task_id_str,
                    statuses=set([Status.CURRENT, Status.IN_PROGRESS, Status.TODO]),
                    exact_match=False,
                )
        else:
            raise

    hunter.workon_task(task.id)
    ctx.invoke(ls, current=True)


def get_estimate(estimate_hours: int | None) -> int:
    """Prompt user for an estimate if one was not given yet."""
    while estimate_hours is None or estimate_hours < 1:
        estimate_hours = typer.prompt("Estimate (hours)", type=int)
        if not estimate_hours or isinstance(estimate_hours, int) and estimate_hours < 1:
            typer.echo("Estimate must be at least 1 hour.")
    return estimate_hours



from unittest import TestCase

from thunter.constants import Status
from thunter.models import Task


class TestTask(TestCase):
    def test_from_db_record(self):
        """Test creating a task from a database record."""
        record = (
            1,
            "Test Task",
            2,
            "This is a test task.",
            "TODO",
            1633036800,
            1633036800,
        )
        task = Task.from_db_record(record)
        self.assertEqual(task.id, 1)
        self.assertEqual(task.name, "Test Task")
        self.assertEqual(task.estimate, 2)
        self.assertEqual(task.description, "This is a test task.")
        self.assertEqual(task.status.value, "TODO")
        self.assertEqual(task.last_modified_at, 1633036800)
        self.assertEqual(task.created_at, 1633036800)
        self.assertEqual(task.estimate_display, "2 hrs")
        self.assertEqual(task.last_modified_at_display, "2021-09-30 21:20:00")

    def test_ordering(self):
        """Test ordering of tasks based on statuses and last modified time."""
        task1 = Task(
            id=1,
            name="Task A",
            estimate=1,
            description="Description A",
            status=Status.TODO,
            last_modified_at=1633011111,
            created_at=1633011111,
        )
        task2 = Task(
            id=2,
            name="Task B",
            estimate=2,
            description="Description B",
            status=Status.IN_PROGRESS,
            last_modified_at=1633022222,
            created_at=1633022222,
        )
        task3 = Task(
            id=3,
            name="Task C",
            estimate=None,
            description="Description C",
            status=Status.FINISHED,
            last_modified_at=1633033333,
            created_at=1633033333,
        )
        task4 = Task(
            id=3,
            name="Task C",
            estimate=None,
            description="Description C",
            status=Status.FINISHED,
            last_modified_at=1633044444,
            created_at=1633044444,
        )

        sorted_tasks = sorted([task4, task3, task2, task1])
        self.assertEqual(sorted_tasks, [task2, task1, task3, task4])



from subprocess import call
import tempfile
from typing import Annotated
import typer

from thunter.cli.ls import ls
from thunter import settings
from thunter.constants import Status
from thunter.settings import thunter_print
from thunter.task_hunter import TaskHunter
from thunter.parser import parse_task_display


app = typer.Typer()


@app.command()
def edit(
    ctx: typer.Context,
    task_identifier: Annotated[
        str | None,
        typer.Argument(
            help="Task ID or name to edit. Defaults to editing the CURRENT task.",
            show_default=False,
        ),
    ] = None,
):
    """Edit a task. Use with caution."""
    hunter = TaskHunter()
    if task_identifier and task_identifier.isdigit():
        task = hunter.get_task(task_identifier)
    else:
        task = hunter.get_task(
            task_identifier=task_identifier,
            statuses={Status.CURRENT, Status.IN_PROGRESS, Status.TODO},
        )

    if not task:
        thunter_print("Could not find task '" + (task_identifier or "CURRENT") + "'")
        return

    with tempfile.NamedTemporaryFile(mode="w", suffix=".tmp") as tf:
        tf.write(hunter.display_task(task.id))
        tf.flush()
        call(settings.EDITOR.split(" ") + [tf.name])

        with open(tf.name, mode="r") as tf:
            tf.seek(0)
            updated_task_to_parse = tf.read()

    parsed_task_data = parse_task_display(updated_task_to_parse)
    hunter.remove_task(task.id)
    new_updated_task = hunter.create_task(
        name=parsed_task_data.name,
        estimate=parsed_task_data.estimate,
        description=parsed_task_data.description,
        status=parsed_task_data.status,
        created_at=task.created_at,
    )
    for history_data in parsed_task_data.history:
        hunter.insert_history(
            taskid=new_updated_task.id,
            is_start=history_data.is_start,
            time=history_data.time,
        )

    ctx.invoke(ls, starts_with=new_updated_task.name, all=True)



from contextlib import contextmanager

import sqlite3

from thunter import settings
from thunter.constants import Status, TableName
from thunter.models.task import Task
from thunter.models.task_history_record import TaskHistoryRecord
from thunter.time import now_sec


class Database:
    """Base class for database interactions, supplying a context manager for
    database connections and handling initialization."""

    def __init__(self, database=None):
        self.database = database or settings.DATABASE

    def init_db(self):
        """Setup the database and tables. Does nothing if the database is already initialized."""
        with self.connect() as con:
            con.execute(
                f"CREATE TABLE IF NOT EXISTS {TableName.TASKS.value} ("
                "id INTEGER PRIMARY KEY AUTOINCREMENT, "
                "name TEXT NOT NULL, "
                "estimate INTEGER, "
                "description TEXT, "
                "status TEXT NOT NULL, "
                "last_modified_at INTEGER NOT NULL DEFAULT (CAST(strftime('%s', current_timestamp) AS integer)), "
                "created_at INTEGER NOT NULL DEFAULT (CAST(strftime('%s', current_timestamp) AS integer))"
                ")"
            )
            con.execute(
                f"CREATE TABLE IF NOT EXISTS {TableName.HISTORY.value} ("
                "id INTEGER PRIMARY KEY AUTOINCREMENT,"
                "taskid INTEGER NOT NULL,"
                "is_start BOOLEAN NOT NULL,"
                "time INTEGER NOT NULL DEFAULT (CAST(strftime('%s', current_timestamp) AS integer)),"
                f"FOREIGN KEY (taskid) REFERENCES {TableName.TASKS.value}(id)"
                ")"
            )
            con.execute(
                "CREATE TRIGGER IF NOT EXISTS last_modified_task_trigger "
                f"AFTER UPDATE ON {TableName.TASKS.value} "
                "BEGIN"
                f"  UPDATE {TableName.TASKS.value} SET last_modified_at = "
                "  (CAST(strftime('%s', current_timestamp) AS integer)) "
                "  WHERE id = NEW.id;"
                "END;"
            )

    @contextmanager
    def connect(self):
        conn = sqlite3.connect(self.database)
        yield conn
        conn.commit()
        conn.close()

    def update_task_field(self, taskid: int, field: str, value: str | int) -> None:
        """Update a specific field of a task in the database."""
        sql = ("UPDATE {table} SET {field}=?, last_modified_at=? WHERE id=?").format(
            table=TableName.TASKS.value, field=field
        )
        sql_params = (value, now_sec(), taskid)
        with self.connect() as conn:
            conn.execute(sql, sql_params)

    def select_from_task(
        self,
        where_clause: str | None = None,
        order_by: str | None = None,
        params: list[str] | None = None,
    ) -> list[Task]:
        return list(
            map(
                Task.from_db_record,
                self.select_from_table(TableName.TASKS, where_clause, order_by, params),
            )
        )

    def select_from_history(
        self,
        where_clause: str | None = None,
        order_by: str | None = None,
        params: list[str] | None = None,
    ) -> list[TaskHistoryRecord]:
        return list(
            map(
                TaskHistoryRecord.from_db_record,
                self.select_from_table(
                    TableName.HISTORY, where_clause, order_by, params
                ),
            )
        )

    def select_from_table(
        self,
        table: TableName,
        where_clause: str | None = None,
        order_by: str | None = None,
        params: list[str] | None = None,
    ):
        sql = "SELECT * FROM {table}".format(table=table.value)
        if where_clause:
            sql += " WHERE " + where_clause
        if order_by:
            sql += " ORDER BY " + order_by

        with self.connect() as conn:
            return conn.execute(sql, params or []).fetchall()

    def insert_task(
        self,
        name: str,
        estimate: int | None,
        description: str | None,
        status: Status,
        created_at: int | None = None,
    ) -> int:
        """Insert a new task into the database and return its ID."""
        if created_at is None:
            sql = (
                f"INSERT INTO {TableName.TASKS.value} "
                "(name,estimate,description,status) "
                "VALUES (?,?,?,?)"
            )
            sql_params = (
                name,
                estimate,
                description,
                status.value,
            )
        else:
            sql = (
                f"INSERT INTO {TableName.TASKS.value} "
                "(name,estimate,description,status,created_at) "
                "VALUES (?,?,?,?,?)"
            )
            sql_params = (
                name,
                estimate,
                description,
                status.value,
                created_at,
            )
        with self.connect() as conn:
            new_task_id = conn.execute(sql, sql_params).lastrowid
            if new_task_id is None:
                raise AssertionError(f"Could not insert task: {name}")
        return new_task_id

    def insert_history(
        self, taskid: int, is_start: bool, time: int | None = None
    ) -> None:
        if time is None:
            sql = (
                f"INSERT INTO {TableName.HISTORY.value} (taskid,is_start) VALUES (?,?)"
            )
            sql_params = (taskid, is_start)
        else:
            sql = ("INSERT INTO {table} (taskid,is_start,time) VALUES (?,?,?)").format(
                table=TableName.HISTORY.value
            )
            sql_params = (taskid, is_start, time)

        with self.connect() as conn:
            conn.execute(sql, sql_params)



from thunter.constants import (
    Status,
    TableName,
    ThunterCouldNotFindTaskError,
    ThunterFoundMultipleTasksError,
)
from thunter.db import Database
from thunter.models import Task, TaskHistoryRecord, TaskIdentifier
from thunter.parser import display_task


class TaskHunter(Database):
    """The tasks manager class for interacting with tasks and their time tracking history."""

    def get_task(
        self,
        task_identifier: TaskIdentifier | None = None,
        statuses: set[Status] | None = None,
        exact_match: bool = False,
    ) -> Task:
        """Fetch a task from the db by its identifier (name or id).

        Defaults to returning the current task if no identifier is given.
        Optionally filter by statuses in the search.

        :param TaskIdentifier task_identifier: The identifier of the task to fetch (name or id).
        :param set[Status] statuses: A set of statuses to filter the task by.
        :param bool exact_match: If True, the task name must match exactly [default: False, will hit with partial matches]

        :return Task: The matching task.

        :raises ThunterCouldNotFindTaskError: If no task is found.
        :raises ThunterFoundMultipleTasksError: If multiple tasks match the identifier.
        """
        if not task_identifier:
            current_task = self.get_current_task()
            if not current_task:
                where_clause = ""
                params = []
                if statuses:
                    where_clause = "status IN (" + ",".join(len(statuses) * "?") + ")"
                    params.extend(map(lambda s: s.value, statuses))
                recent_tasks = self.select_from_task(
                    order_by="last_modified_at DESC",
                    where_clause=where_clause,
                    params=params,
                )
                if recent_tasks:
                    return recent_tasks[0]
                raise ThunterCouldNotFindTaskError("No task found.")
            return current_task

        params: list[str] = []
        if isinstance(task_identifier, int) or task_identifier.isdigit():
            where_clause = "id=?"
            params = [str(task_identifier)]
        elif exact_match:
            where_clause = "name = ?"
            params = [task_identifier]
        else:
            where_clause = "name LIKE ?"
            params = [task_identifier + "%"]

        if statuses:
            where_clause += " AND status IN (" + ",".join(len(statuses) * "?") + ")"
            params.extend(map(lambda s: s.value, statuses))

        order_by: str = "last_modified_at DESC"
        tasks = self.select_from_task(
            where_clause=where_clause, order_by=order_by, params=params
        )

        if len(tasks) == 0:
            raise ThunterCouldNotFindTaskError(
                f"Could not find task for identifier: {task_identifier}"
            )
        elif len(tasks) > 1:
            raise ThunterFoundMultipleTasksError(
                f"Found multiple tasks for identifier: {task_identifier}"
            )

        return tasks[0]

    def display_task(self, taskid: int):
        """Return a string representation of the task and its history.

        This representation is able to be parsed and is used by the edit command."""
        task = self.get_task(taskid)
        task_history = self.get_history([taskid])
        return display_task(task=task, task_history=task_history)

    def create_task(
        self,
        name: str,
        estimate: int | None = None,
        description: str | None = None,
        status: Status = Status.TODO,
        created_at: int | None = None,
    ) -> Task:
        """Create a new task in the database and return it.

        :param name: The name of the task.
        :param estimate: The estimated time for the task in hours.
        :param description: A description of the task.
        :param status: The initial status of the task [default: TODO].
        """
        if name.isdigit():
            raise ValueError(
                "Task cannot be a number, as that would conflict with task IDs."
            )
        new_task_id = self.insert_task(
            name=name,
            estimate=estimate,
            description=description,
            status=status,
            created_at=created_at,
        )
        return self.get_task(new_task_id)

    def get_tasks(
        self,
        statuses: set[Status] | None = None,
        starts_with: str | None = None,
        contains: str | None = None,
    ) -> list[Task]:
        """Fetch a list tasks from the database with optional filters."""
        where_clause_param_pairs = []
        if starts_with:
            where_clause_param_pairs.append(("name LIKE ?", (starts_with + "%",)))
        if contains:
            where_clause_param_pairs.append(("name LIKE ?", ("%" + contains + "%",)))
        if statuses:
            where_clause_param_pairs.append(
                (
                    "status IN (" + ",".join(len(statuses) * "?") + ")",
                    map(lambda s: s.value, statuses),
                )
            )
        if where_clause_param_pairs:
            where_clauses, where_params = zip(*where_clause_param_pairs)
            where_clause = " AND ".join(where_clauses)
            params = [param for params in where_params for param in params]
        else:
            where_clause = None
            params = None

        tasks = self.select_from_task(where_clause=where_clause, params=params)
        return sorted(tasks)

    def get_history(self, taskids: list[int]) -> list[TaskHistoryRecord]:
        """Fetch the history records for a given task or list of tasks."""
        assert all(map(lambda taskid: isinstance(taskid, int), taskids))

        where_clause = "taskid IN (" + ",".join(len(taskids) * "?") + ")"
        history = self.select_from_history(
            where_clause=where_clause, params=list(map(str, taskids))
        )
        return sorted(history)

    def get_current_task(self) -> Task | None:
        """Fetch the current task from the database.

        There should only ever be one current task actively being worked on at
        a time.
        """
        current_tasks = self.select_from_task(
            where_clause="status IN (?)",
            params=[Status.CURRENT.value],
        )
        if len(current_tasks) == 0:
            return None
        elif len(current_tasks) > 1:
            raise AssertionError("More than one current task? How!?")

        return current_tasks[0]

    def workon_task(self, task_identifier: TaskIdentifier) -> None:
        """Start working on a task by its identifier (name or id).

        Stops tracking time spent on the current task if there is one, changing its status
        to IN_PROGRESS, and starts tracking time on the new task, changing its status to CURRENT.

        No-op if the identifier matches the current task.
        """
        task = self.get_task(task_identifier)
        current_task = self.get_current_task()
        if current_task:
            if current_task.id == task.id:
                return
            self.insert_history(taskid=current_task.id, is_start=False)
            self.update_task_field(current_task.id, "status", Status.IN_PROGRESS.value)
        self.insert_history(taskid=task.id, is_start=True)
        self.update_task_field(task.id, "status", Status.CURRENT.value)

    def stop_current_task(self) -> Task | None:
        """Stop tracking the current task and change its status to IN_PROGRESS."""
        current_task = self.get_current_task()
        if not current_task:
            return
        self.insert_history(taskid=current_task.id, is_start=False)
        self.update_task_field(current_task.id, "status", Status.IN_PROGRESS.value)
        return self.get_task(current_task.id)

    def finish_task(self, taskid: int) -> None:
        """Mark a task as finished and update its status."""
        task = self.get_task(taskid)
        if task.status == Status.CURRENT:
            self.insert_history(taskid=task.id, is_start=False)

        if task.status != Status.FINISHED:
            self.update_task_field(
                taskid=task.id, field="status", value=Status.FINISHED.value
            )

    def estimate_task(self, taskid: int, estimate: int) -> None:
        """(Re)set the estimate for a task."""
        self.update_task_field(taskid, "estimate", estimate)

    def remove_task(self, taskid: int) -> None:
        """Permanently delete a task and its history from the database."""
        delete_task_sql = "DELETE from {table} WHERE id=?".format(
            table=TableName.TASKS.value
        )
        delete_history_sql = "DELETE from {table} WHERE taskid=?".format(
            table=TableName.HISTORY.value
        )
        with self.connect() as conn:
            conn.execute(delete_task_sql, [taskid])
            conn.execute(delete_history_sql, [taskid])



from thunter.tests import CliCommandTestBaseClass
from thunter.cli import thunter_cli_app


class TestRm(CliCommandTestBaseClass):
    def test_rm_unsure(self):
        result = self.runner.invoke(
            thunter_cli_app, ["rm", "a finished task"], input="n\n"
        )
        self.assertIn("Didn't remove a finished task.", result.output)

    def test_rm_sure(self):
        result = self.runner.invoke(
            thunter_cli_app, ["rm", "a finished task"], input="y\n"
        )
        self.assertIn("Removed a finished task!", result.output)

    def test_rm_force(self):
        result = self.runner.invoke(
            thunter_cli_app, ["rm", "a finished task", "--force"]
        )
        self.assertIn("Removed a finished task!", result.output)



from unittest import TestCase

from thunter.constants import Status
from thunter.db import Database
from thunter.models.task import Task
from thunter.models.task_history_record import TaskHistoryRecord
from thunter.tests import setUpTestDatabase, tearDownTestDatabase


class TestTaskHunter(TestCase):
    def setUp(self):
        self.env = setUpTestDatabase()
        self.database = Database()

    def tearDown(self):
        """Remove the temporary environment and database after tests."""
        tearDownTestDatabase(self.env)

    def test__init__(self):
        thunter = Database()
        self.assertEqual(thunter.database, self.env.DATABASE)

        thunter = Database("/database.db")
        self.assertEqual(thunter.database, "/database.db")

    def test_select_from_history(self):
        history = self.database.select_from_history(
            where_clause="taskid = ?",
            order_by="time ASC",
            params=["1"],
        )
        expected_history = [
            TaskHistoryRecord(id=3, taskid=1, is_start=True, time=1722208103),
            TaskHistoryRecord(id=4, taskid=1, is_start=False, time=1753732126),
        ]
        self.assertEqual(history, expected_history)

    def test_select_from_task(self):
        history = self.database.select_from_task(
            where_clause="id = ?",
            order_by="id ASC",
            params=["1"],
        )
        expected_history = [
            Task(
                id=1,
                name="a test task",
                estimate=4,
                description=None,
                status=Status.IN_PROGRESS,
                last_modified_at=1753732126,
                created_at=1753732126,
            ),
        ]
        self.assertEqual(history, expected_history)



import typer
from IPython import embed

from thunter.constants import Status
from thunter.analyzer import TaskAnalyzer

app = typer.Typer()


@app.command()
def analyze():
    """Create a new task."""
    analyzer = TaskAnalyzer()
    df = analyzer.fetch_data_df()  # noqa: F841
    tasks_df = analyzer.fetch_tasks_df()
    history_df = analyzer.fetch_history_df()

    # TODO: merge dataframes and filter to just FINISHED tasks
    # Convert start times to negative
    history_df.loc[history_df["is_start"], "time"] *= -1
    # sum up the time for each task
    actual_time = history_df.groupby("taskid", sort=False)["time"].sum()
    finished_tasks = tasks_df[tasks_df["status"] == Status.FINISHED.value]
    comparison_df = finished_tasks.merge(actual_time, left_on="id", right_on="taskid")  # noqa: F841

    # TODO: time series plot of estimate vs actual

    embed()



from typing import Annotated
import typer

from thunter.constants import Status
from thunter.settings import thunter_print
from thunter.task_hunter import TaskHunter


app = typer.Typer()


@app.command("f, finish")
def finish(task_id: Annotated[str | None, typer.Argument()] = None):
    """Finish a task (defaults to finish current task)."""
    hunter = TaskHunter()
    task = hunter.get_task(task_id)
    if task.status == Status.CURRENT:
        hunter.stop_current_task()

    hunter.finish_task(task.id)
    thunter_print(f"Finished [green]{task.name}[/green]!")



from subprocess import call
import typer

from thunter.settings import DATABASE


app = typer.Typer()


@app.command()
def db():
    """Access the sqlite database directly. Tables: tasks, history

    sqlite3> SELECT * FROM tasks

    sqlite3> SELECT * FROM history WHERE taskid = 1 ORDER BY time DESC
    """
    call(["sqlite3", DATABASE])



from thunter.tests import CliCommandTestBaseClass
from thunter.cli import thunter_cli_app


class TestCreate(CliCommandTestBaseClass):
    def test_create_task(self):
        result = self.runner.invoke(
            thunter_cli_app,
            ["create", "Test Task", "--estimate", "2", "--description", "A test task"],
        )
        self.assertIn("Test Task", result.output)
        self.assertIn("2", result.output)
        self.assertIn("A test task", result.output)

    def test_create_with_estimate_prompt(self):
        result = self.runner.invoke(
            thunter_cli_app,
            ["create", "Test Task with Prompt"],
            input="3\n",
        )
        self.assertIn("Test Task with Prompt", result.output)
        self.assertIn("3", result.output)



import typer

from thunter.settings import thunter_print
from thunter.task_hunter import TaskHunter


app = typer.Typer()


@app.command()
def stop():
    """Stop working on current task."""
    hunter = TaskHunter()
    stopped_task = hunter.stop_current_task()
    if not stopped_task:
        thunter_print("No current task to stop.", style="yellow")
    else:
        thunter_print(f"Stopped working on [yellow]{stopped_task.name}[/yellow]!")



from enum import Enum


class TableName(Enum):
    TASKS = "tasks"
    HISTORY = "history"


class Status(Enum):
    CURRENT = "Current"
    IN_PROGRESS = "In Progress"
    TODO = "TODO"
    FINISHED = "Finished"


STATUS_ORDERING = [
    Status.CURRENT.value,
    Status.IN_PROGRESS.value,
    Status.TODO.value,
    Status.FINISHED.value,
]

TIME_FORMAT = "%Y-%m-%d %H:%M:%S"


class ThunterError(Exception):
    exit_status = 1


class ThunterCouldNotFindTaskError(ThunterError):
    exit_status = 2


class ThunterAlreadyWorkingOnTaskError(ThunterError):
    exit_status = 3


class ThunterNoCurrentTaskError(ThunterError):
    exit_status = 4


class ThunterFoundMultipleTasksError(ThunterError):
    exit_status = 5


class ThunterTaskValidationError(ThunterError):
    exit_status = 6


class ThunterNotInitializedError(ThunterError):
    exit_status = 7



from typing import Annotated
import typer

from thunter.settings import thunter_print
from thunter.task_hunter import TaskHunter


app = typer.Typer()


@app.command()
def show(task_id: Annotated[str | None, typer.Argument()] = None):
    """Display task. Defaults to the currently active task if there is one."""
    hunter = TaskHunter()
    task = hunter.get_task(task_id)
    thunter_print(hunter.display_task(task.id))



from unittest import TestCase

from thunter.constants import (
    Status,
    ThunterCouldNotFindTaskError,
    ThunterFoundMultipleTasksError,
)
from thunter.task_hunter import TaskHunter
from thunter.tests import setUpTestDatabase, tearDownTestDatabase


class TestTaskHunter(TestCase):
    def setUp(self):
        self.env = setUpTestDatabase()
        self.thunter = TaskHunter()

    def tearDown(self):
        """Remove the temporary environment and database after tests."""
        tearDownTestDatabase(self.env)

    def test_get_task_by_id(self):
        task_by_id = self.thunter.get_task(1)
        self.assertEqual(task_by_id.name, "a test task")
        self.assertEqual(task_by_id.status, Status.IN_PROGRESS)

        with self.assertRaises(ThunterCouldNotFindTaskError):
            self.thunter.get_task(-1)

    def test_get_task_by_name(self):
        task_by_name = self.thunter.get_task("a test task")
        self.assertEqual(task_by_name.id, 1)
        self.assertEqual(task_by_name.name, "a test task")
        self.assertEqual(task_by_name.status, Status.IN_PROGRESS)

        with self.assertRaises(ThunterCouldNotFindTaskError):
            self.thunter.get_task("nonexistent task")

        with self.assertRaises(ThunterFoundMultipleTasksError):
            self.thunter.get_task("identically named task")

    def test_get_current_task(self):
        default_current_task = self.thunter.get_task()
        self.assertEqual(default_current_task.id, 5)
        self.assertEqual(default_current_task.name, "a long task")
        self.assertEqual(default_current_task.status, Status.CURRENT)

        current_task = self.thunter.get_current_task()
        self.assertEqual(current_task, default_current_task)

        stopped_current_task = self.thunter.stop_current_task()
        self.assertEqual(stopped_current_task.id, current_task.id)  # type: ignore
        self.assertEqual(self.thunter.get_current_task(), None)

        default_recent_task = self.thunter.get_task()
        self.assertEqual(default_recent_task.id, default_current_task.id)

    def test_get_finished_task(self):
        finished_task = self.thunter.get_task("a finished task")
        self.assertEqual(finished_task.id, 4)
        self.assertEqual(finished_task.name, "a finished task")
        self.assertEqual(finished_task.status, Status.FINISHED)

        with self.assertRaises(ThunterCouldNotFindTaskError):
            self.thunter.get_task(finished_task.id, statuses={Status.TODO})

    def test_display_task(self):
        task_display = self.thunter.display_task(1)
        self.assertEqual(
            task_display,
            "NAME: a test task\n"
            "ESTIMATE: 4\n"
            "STATUS: In Progress\n"
            "DESCRIPTION: None\n"
            "\n"
            "HISTORY\n"
            "Start\t2025-07-28 19:48:23\n"
            "Stop\t2025-07-28 19:48:46\n",
        )

    def test_create_task(self):
        new_task = self.thunter.create_task(
            name="New Task",
            estimate=5,
            description="A new task for testing.",
            status=Status.TODO,
        )
        self.assertEqual(new_task.name, "New Task")
        self.assertEqual(new_task.estimate, 5)
        self.assertEqual(new_task.description, "A new task for testing.")
        self.assertEqual(new_task.status, Status.TODO)

        # Verify the task was added to the database
        fetched_task = self.thunter.get_task(new_task.id)
        self.assertEqual(fetched_task, new_task)

    def test_create_task_invalid_name(self):
        with self.assertRaises(ValueError) as exception:
            self.thunter.create_task(
                name="123",
                estimate=5,
                description="Invalid number-only task name.",
                status=Status.TODO,
            )
        self.assertEqual(
            str(exception.exception),
            "Task cannot be a number, as that would conflict with task IDs.",
        )

    def test_get_tasks(self):
        tasks = self.thunter.get_tasks()
        self.assertEqual(len(tasks), 6)
        self.assertIn("a test task", [task.name for task in tasks])
        self.assertIn("a finished task", [task.name for task in tasks])
        self.assertIn("a long task", [task.name for task in tasks])

        starts_with_a_tasks = self.thunter.get_tasks(starts_with="a")
        self.assertEqual(len(starts_with_a_tasks), 4)
        self.assertIn("a test task", [task.name for task in starts_with_a_tasks])
        self.assertIn("a finished task", [task.name for task in starts_with_a_tasks])

        contains_great_task = self.thunter.get_tasks(contains="great")
        self.assertEqual(len(contains_great_task), 1)
        self.assertEqual(contains_great_task[0].name, "another great test task")

        todo_tasks = self.thunter.get_tasks(statuses={Status.TODO})
        self.assertEqual(len(todo_tasks), 3)
        self.assertEqual(
            [
                "another great test task",
                "identically named task",
                "identically named task",
            ],
            sorted([task.name for task in todo_tasks]),
        )

    def test_workon_task(self):
        current_task = self.thunter.get_task()
        self.assertEqual(current_task.name, "a long task")
        self.assertEqual(current_task.status, Status.CURRENT)
        current_task_history = self.thunter.get_history([current_task.id])

        next_task = self.thunter.get_task(task_identifier="a test task")
        next_task_history = self.thunter.get_history([next_task.id])

        # Start working on the next task
        self.thunter.workon_task(next_task.id)

        new_current_task = self.thunter.get_task()
        self.assertEqual(new_current_task, next_task)

        previously_current_task = self.thunter.get_task(current_task.id)
        previously_current_task_history = self.thunter.get_history(
            [previously_current_task.id]
        )
        new_current_task_history = self.thunter.get_history([new_current_task.id])

        self.assertEqual(len(new_current_task_history), len(next_task_history) + 1)
        self.assertEqual(
            len(previously_current_task_history), len(current_task_history) + 1
        )
        self.assertEqual(new_current_task.status, Status.CURRENT)
        self.assertEqual(previously_current_task.status, Status.IN_PROGRESS)

        # Already working on this task, should be a no-op
        self.thunter.workon_task(1)

        unchanged_current_task = self.thunter.get_task()
        unchanged_current_task_history = self.thunter.get_history(
            [unchanged_current_task.id]
        )
        self.assertEqual(unchanged_current_task, new_current_task)
        self.assertEqual(unchanged_current_task.status, new_current_task.status)
        self.assertEqual(unchanged_current_task_history, new_current_task_history)

    def test_stop_current_task(self):
        current_task = self.thunter.get_current_task()
        self.assertIsNotNone(current_task)
        assert current_task

        current_history = self.thunter.get_history([current_task.id])
        self.assertEqual(current_task.status, Status.CURRENT)
        self.assertEqual(len(current_history), 3)

        stopped_task = self.thunter.stop_current_task()
        assert stopped_task

        self.assertEqual(stopped_task.id, current_task.id)
        self.assertEqual(stopped_task.status, Status.IN_PROGRESS)

        # Verify the task is no longer current
        self.assertIsNone(self.thunter.get_current_task())

        # Verify the task history was updated
        stopped_history = self.thunter.get_history([stopped_task.id])
        self.assertEqual(len(stopped_history), len(current_history) + 1)

        # stopping the current task again should not change anything
        self.assertIsNone(self.thunter.stop_current_task())

    def test_update_finish_task(self):
        current_task = self.thunter.get_task()
        current_history = self.thunter.get_history([current_task.id])
        self.assertEqual(current_task.status, Status.CURRENT)

        self.thunter.finish_task(current_task.id)

        finished_task = self.thunter.get_task(current_task.id)
        self.assertEqual(finished_task.status, Status.FINISHED)

        # Verify the task history was updated
        finished_history = self.thunter.get_history([finished_task.id])
        self.assertEqual(len(finished_history), len(current_history) + 1)

    def test_estimate_task(self):
        current_task = self.thunter.get_task()
        self.assertEqual(current_task.estimate, 32)

        self.thunter.estimate_task(taskid=current_task.id, estimate=10)

        updated_task = self.thunter.get_task(current_task.id)
        self.assertEqual(updated_task.estimate, 10)

    def test_remove_task(self):
        current_task = self.thunter.get_task()
        self.thunter.remove_task(current_task.id)
        with self.assertRaises(ThunterCouldNotFindTaskError):
            self.thunter.get_task(current_task.id)



import calendar
from dataclasses import dataclass
from time import strptime
from typing import TypeGuard

import pyparsing as pp

from thunter.constants import TIME_FORMAT, ThunterTaskValidationError, Status
from thunter.models import Task, TaskHistoryRecord


@dataclass
class ParsedTaskHistoryRecord:
    is_start: bool
    time: int


@dataclass
class ParsedTaskData:
    name: str
    estimate: int | None
    description: str | None
    status: Status
    history: list[ParsedTaskHistoryRecord]


def parse_time(t: pp.ParseResults) -> int:
    """parser action for converting time string to seconds"""
    if isinstance(t.time, str):
        return calendar.timegm(strptime(t.time, TIME_FORMAT))
    raise ValueError


word = pp.Word(pp.alphanums + " .!?&-_()/\\")
number = pp.common.integer
status_type = pp.one_of([s.value for s in Status]).set_parse_action(
    lambda t: Status(t.status)
)
is_start = pp.one_of(["Start", "Stop"])("is_start").set_parse_action(
    lambda t: t[0] == "Start"
)
time = pp.Word(pp.nums + " -:")("time").set_parse_action(parse_time)
name = pp.Suppress("NAME:") + word("name")
estimate = pp.Suppress("ESTIMATE:") + number("estimate")
status = pp.Suppress("STATUS:") + status_type("status")


description = pp.Suppress("DESCRIPTION:") + pp.rest_of_line(
    "description"
).set_parse_action(
    lambda t: t.description.strip() if isinstance(t.description, str) else ""
)

history_record = pp.Group(is_start("is_start") + time("time"))
history = pp.Suppress(pp.Keyword("HISTORY")) + pp.ZeroOrMore(history_record)("history")

task = name + estimate + status + description + history


def is_parsed_history(history) -> TypeGuard[list[ParsedTaskHistoryRecord]]:
    return all(
        history_record.is_start in [True, False]
        and isinstance(history_record.time, int)
        for history_record in history
    )


def is_parsed_task(data) -> TypeGuard[ParsedTaskData]:
    return (
        isinstance(data.name, str)
        and isinstance(data.estimate, (int, type(None)))
        and isinstance(data.description, (str, type(None)))
        and isinstance(data.status, Status)
        and is_parsed_history(data.history)
    )


def parse_task_display(task_display: str) -> ParsedTaskData:
    """Parses a task's display string into the data necessary to create the task and it's history"""
    task_data = task.parse_string(task_display)

    if not (is_parsed_task(task_data)):
        raise AssertionError("Invalid parsed task data")

    validate_task_data(task_data)
    return task_data


def thunter_assert(expr, message):
    if not expr:
        error_message = f"[red]Task Validation Error:[/red] {message}"
        raise ThunterTaskValidationError(error_message)


def validate_task_data(task_data: ParsedTaskData) -> None:
    if task_data.status == Status.TODO:
        thunter_assert(
            len(task_data.history) == 0, "Can't have a history if the status is TODO"
        )
    else:
        thunter_assert(
            len(task_data.history) > 0,
            "Must have a history if status is %s" % task_data.status.value,
        )

    if task_data.status == Status.CURRENT:
        last_history_record = task_data.history[-1]
        thunter_assert(
            last_history_record.is_start,
            "Last history record must be a Start if the status is Current",
        )
    elif task_data.status in [Status.IN_PROGRESS, Status.FINISHED]:
        last_history_record = task_data.history[-1]
        thunter_assert(
            not last_history_record.is_start,
            "Last history record must be a Stop if the status is %s"
            % task_data.status.value,
        )

    expect_start = True
    last_history_time = 0
    for history_data in task_data.history:
        thunter_assert(
            last_history_time <= history_data.time,
            "History must be in ascending order by time",
        )
        thunter_assert(
            history_data.is_start == expect_start,
            "History must alternate between Start and Stop",
        )
        expect_start = not expect_start
        last_history_time = history_data.time


def display_task(task: Task, task_history: list[TaskHistoryRecord]) -> str:
    """Displays a task in a human-readable and parser friendly format."""
    lines = []
    lines.append("NAME: %s" % task.name)
    lines.append("ESTIMATE: %s" % task.estimate)
    lines.append("STATUS: %s" % task.status.value)
    lines.append("DESCRIPTION: %s" % task.description)
    lines.append("")
    lines.append("HISTORY")
    for history_record in task_history:
        record_type = "Start" if history_record.is_start else "Stop"
        lines.append(record_type + "\t" + history_record.time_display)
    return "\n".join(lines + [""])



from unittest import TestCase

import pandas as pd

from thunter.analyzer import TaskAnalyzer
from thunter.tests import setUpTestDatabase, tearDownTestDatabase


class TestTaskAnalyzer(TestCase):
    def setUp(self):
        self.env = setUpTestDatabase()
        self.analyzer = TaskAnalyzer()

    def tearDown(self):
        """Remove the temporary environment and database after tests."""
        tearDownTestDatabase(self.env)

    def test_fetch_data_df(self):
        df = self.analyzer.fetch_data_df()
        self.assertIsInstance(df, pd.DataFrame)
        self.assertIn("estimate", df.columns)
        self.assertIn("is_start", df.columns)
        self.assertIn("time", df.columns)
        self.assertEqual(df["estimate"].dtype, "Int32")
        self.assertEqual(df["is_start"].dtype, "boolean")
        self.assertEqual(df["time"].dtype, "Int64")



import re
from thunter.tests import CliCommandTestBaseClass
from thunter.cli import thunter_cli_app


class TestRestart(CliCommandTestBaseClass):
    def test_restart_finished_task(self):
        result = self.runner.invoke(thunter_cli_app, ["restart", "a finished task"])
        self.assertTrue(re.search(r"a finished task.*Current", result.output))

    def test_restart_no_task(self):
        result = self.runner.invoke(thunter_cli_app, ["restart", "a long task"])
        self.assertIn(
            "Could not find task for identifier: a long task", str(result.exception)
        )
        self.assertGreater(result.exit_code, 0)



import re
from thunter.tests import CliCommandTestBaseClass
from thunter.cli import thunter_cli_app


class TestWorkon(CliCommandTestBaseClass):
    def test_workon_task(self):
        result = self.runner.invoke(thunter_cli_app, ["show"])
        self.assertIn("a long task", result.output)
        self.assertNotIn("a test task", result.output)

        self.runner.invoke(thunter_cli_app, ["workon", "a test task"])
        result = self.runner.invoke(thunter_cli_app, ["show"])
        self.assertNotIn("a long task", result.output)
        self.assertIn("a test task", result.output)

    def test_workon_and_create_task(self):
        result = self.runner.invoke(
            thunter_cli_app, ["workon", "a brand new task", "--create"], input="1\n"
        )
        self.assertTrue(re.search("a brand new task.*Current", result.output))

    def test_workon_nonexistent_task_by_id(self):
        result = self.runner.invoke(thunter_cli_app, ["workon", "999"])
        self.assertIn("Could not find task for identifier: 999", str(result.exception))
        self.assertGreater(result.exit_code, 0)

    def test_workon_nonexistent_task_by_name(self):
        result = self.runner.invoke(thunter_cli_app, ["workon", "a nonexistent task"])
        self.assertIn(
            "Could not find task for identifier: a nonexistent task",
            str(result.exception),
        )
        self.assertGreater(result.exit_code, 0)

    def test_workon_partial_task_id_match(self):
        result = self.runner.invoke(thunter_cli_app, ["workon", "another"])
        self.assertIn("another great test task", result.output)
        self.assertNotIn("a long task", result.output)

    def test_workon_last_modified_task(self):
        result = self.runner.invoke(thunter_cli_app, ["stop"])
        self.assertIn("a long task", result.output)

        result = self.runner.invoke(thunter_cli_app, ["workon"])
        self.assertIn("a long task", result.output)

    def test_no_tasks_to_workon(self):
        self.runner.invoke(thunter_cli_app, ["finish", "1"])
        self.runner.invoke(thunter_cli_app, ["finish", "2"])
        self.runner.invoke(thunter_cli_app, ["finish", "5"])
        self.runner.invoke(thunter_cli_app, ["finish", "6"])
        self.runner.invoke(thunter_cli_app, ["finish", "7"])

        result = self.runner.invoke(thunter_cli_app, ["workon"])
        self.assertGreater(result.exit_code, 0)
        self.assertIn("No task found", str(result.exception))

    def test_workon_create_keeps_prompting_for_estimate(self):
        result = self.runner.invoke(
            thunter_cli_app,
            ["workon", "a brand new task", "--create"],
            input="0\n2\n",
        )
        self.assertTrue(re.search("Estimate must be at least 1 hour", result.output))
        self.assertTrue(re.search("a brand new task.*Current", result.output))
        self.assertTrue(re.search("a brand new task.*2 hrs.*Current", result.output))



from datetime import datetime
from time import gmtime, strftime

from thunter.constants import TIME_FORMAT


def now_sec() -> int:
    """Returns the current time as integer seconds since epoch."""
    return int(datetime.now().timestamp())


def display_time(seconds: int) -> str:
    """Formats seconds into a datetime string."""
    return strftime(TIME_FORMAT, gmtime(seconds))



from typing import Annotated
import typer

from thunter.constants import ThunterCouldNotFindTaskError
from thunter.settings import thunter_print
from thunter.task_hunter import TaskHunter


app = typer.Typer()


@app.command()
def estimate(
    estimate: Annotated[int, typer.Argument()],
    task_identifier: Annotated[
        str | None,
        typer.Option(
            "--task-identifier",
            "-t",
            help="Estimate task by ID or name, instead of current task",
            show_default=False,
        ),
    ] = None,
):
    """Estimate how long a task will take."""
    hunter = TaskHunter()
    if task_identifier:
        task = hunter.get_task(task_identifier)
    else:
        task = hunter.get_current_task()

    if not task:
        raise ThunterCouldNotFindTaskError(
            "No current task found to estimate. Use --task-identifier / -t to specify a task."
        )

    hunter.estimate_task(task.id, estimate)
    task = hunter.get_task(task.id)
    thunter_print(
        f"[green]{task.name}[/green] estimated to take [yellow]{task.estimate_display}[/yellow]."
    )



import os
import shutil
from typing import Annotated

import typer

from thunter import settings
from thunter.db import Database
from thunter.settings import thunter_print, needs_init

app = typer.Typer()


@app.command()
def init(
    force: Annotated[
        bool, typer.Option("--force", "-f", help="Skip confirmation prompt")
    ] = False,
):
    """Initialize thunter sqlite database.

    The database file name can be set with THUNTER_DATABASE_NAME, defaults to 'database.db'.
    The database file can be found in the THUNTER_DIRECTORY, defaults to ~/.thunter.
    """
    thunter_print("Initializing THunter...")
    if not needs_init():
        prompt = "WARNING: Are you sure you want to re-initialize? You will lose all tasks and tracking info [yN]"
        user_sure = force or input(prompt).lower() == "y"
        if not user_sure:
            thunter_print("Aborting re-initialization")
            raise typer.Exit()
        thunter_print(f"Deleting THunter directory: {settings.THUNTER_DIR}")
        shutil.rmtree(settings.THUNTER_DIR)

    if not os.path.exists(settings.THUNTER_DIR):
        thunter_print(f"Creating THunter directory: {settings.THUNTER_DIR}")
        os.mkdir(settings.THUNTER_DIR)

    thunter_print(f"Creating sqlite database {settings.DATABASE}")
    db = Database()
    db.init_db()

    thunter_print("THunter initialized successfully!")
